#ifndef ARBUTILS___OPTIONALSCOPEDPTR_HPP
#define ARBUTILS___OPTIONALSCOPEDPTR_HPP

#include "../Ensure.hpp"

namespace ArbUt {
    /// @brief An optional unique pointer is used to indicate a pointer that is owned by its holder, and will be deleted
    /// when its owner is deleted.
    /// @details A unique pointer is used to indicate a pointer that is owned by an object, and that needs to be deleted
    /// when its owner is deleted.
    template <class T> class OptionalScopedPtr {
    private:
        T* nullable _raw;

    public:
        /// @brief Initialise a ScopedPtr with a specific raw pointer.
        inline OptionalScopedPtr<T>(T* nullable ptr) : _raw(ptr){};
        /// @brief Initialise a ScopedPtr from a copy.
        inline OptionalScopedPtr<T>(const OptionalScopedPtr<T>& other) : _raw(other._raw){};
        /// @brief Initialise a ScopedPtr with a std unique_ptr.
        inline OptionalScopedPtr<T>(const std::unique_ptr<T>& other) : _raw(other.get()){};

        ~OptionalScopedPtr() noexcept { delete _raw; }

        /// @brief Copy operator.
        inline OptionalScopedPtr<T>& operator=(const OptionalScopedPtr<T>& rhs) {
            if (this == &rhs) {
                return *this;
            }
            _raw = rhs._raw;
            return *this;
        }

        /// @brief Assign operator with raw pointer.
        inline OptionalScopedPtr<T>& operator=(T* nullable rhs) {
            if (_raw == rhs) {
                return *this;
            }
            _raw = rhs;
            return *this;
        }

        /// @brief Return whether the pointer is null or not.
        [[nodiscard]] inline bool HasValue() const noexcept { return _raw != nullptr; }

        /// @brief Get the raw underlying pointer, and take ownership of it. This removes the existing pointer in here.
        inline T* nullable TakeOwnership() noexcept {
            auto raw = _raw;
            _raw = nullptr;
            return raw;
        }
        /// @brief Get the raw underlying pointer, without taking ownership of it.
        inline T* nullable GetValue() noexcept { return _raw; }

        /// @brief Check equality of two ScopedPtr objects
        inline bool operator==(const OptionalScopedPtr& rhs) const noexcept { return _raw == rhs._raw; }
        /// @brief Check equality of pointers
        inline bool operator==(T* nullable rhs) const noexcept { return _raw == rhs; }
        /// @brief Check equality of two ScopedPtr objects
        inline bool operator!=(const OptionalScopedPtr& rhs) const noexcept { return _raw != rhs._raw; }
        /// @brief Check equality of pointers
        inline bool operator!=(T* nullable rhs) const noexcept { return _raw != rhs; }
        /// @brief Implicit cast to retrieve optional borrowed pointer.
        inline operator OptionalBorrowedPtr<T>() const noexcept { return _raw; }
    };
}

namespace std {
    /// @brief Helper class for allowing hashing of OptionalScopedPtr.
    template <class T> struct hash<ArbUt::OptionalScopedPtr<T>> {
        /// @brief Returns a hash of for an optional scoped Pointer. Effectively just the raw memory address.
        /// @param k An optional scoped pointer.
        /// @return The hash of the optional scoped pointer.
        std::size_t operator()(const ArbUt::OptionalScopedPtr<T>& k) const { return (size_t)k.GetRaw(); }
    };
}

#endif // ARBUTILS___OPTIONALSCOPEDPTR_HPP
